using System;
using UnityEngine;
using UnityEngine.Events;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using SimpleJSON;
using System.Linq;
using MVR;
using MeshVR;
using MVR.FileManagementSecure;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using GPUTools.Cloth.Scripts;
using Unity.Collections;
using System.Diagnostics;

namespace Stopper {
	public class Rails : MVRScript {
		
        string selectedAnimationPattern;
        string selectedController;

        Dictionary<string,FreeControllerV3> controllers = new Dictionary<string, FreeControllerV3>();

        JSONStorableStringChooser animationSelection;
        JSONStorableStringChooser controllerSelection;
        JSONStorableBool hardLink;
        JSONStorableBool linkRotation;

        JSONStorableFloat linkJointSpring;
        JSONStorableFloat linkMaxForce;

        JSONStorableFloat maxVelocity;

        JSONStorableFloat existingLinkType;

        Atom railDriver;
        AnimationPattern animationPattern;
        
        FreeControllerV3 controller;
        FreeControllerV3 railDriverController;

        Rigidbody linkedRB;

        ConfigurableJoint controllerJoint;
        ConfigurableJoint railJoint;
        JointDrive railJointDrive = new JointDrive {
            maximumForce = 100000f,
            positionSpring = 100000f,
            positionDamper = 0f
        };
        
        float previousClosestCurvePosition = 0f;
        bool rotationLinked = false;

		List<UnityEngine.Object> createdObjects = new List<UnityEngine.Object>();

        public class LinkType {
            public static float NONE = 0f;
            public static float RIGIDBODY = 1f;
            public static float CONTROLLER = 2f;
        }

		public override void Init() {

            foreach(var controller in containingAtom.freeControllers){
                controllers[controller.name] = controller;
            }

            animationSelection = new JSONStorableStringChooser("animationSelection", GetAnimationPatterns(), null, "Animation Pattern", (string animationUid) => {
                selectedAnimationPattern = animationUid;
			});
			
			RegisterStringChooser(animationSelection);
			var popup = CreateScrollablePopup(animationSelection, false);
			popup.popup.onOpenPopupHandlers += () => {
				animationSelection.choices = GetAnimationPatterns();
			};

            controllerSelection = new JSONStorableStringChooser("controllerSelection", controllers.Keys.ToList(), null, "Controller", (string controllerName) => {
                selectedController = controllerName;
			});

            RegisterStringChooser(controllerSelection);
			CreateScrollablePopup(controllerSelection, false);

            linkRotation = new JSONStorableBool("Link Rotation", false);
            RegisterBool(linkRotation);
            var linkRotationToggle = CreateToggle(linkRotation, false);

            var linkBodyButton = CreateButton("Link Body");
            linkBodyButton.button.onClick.AddListener(() =>
            {
                StartCoroutine("LinkBody");
            });

            var linkControllerButton = CreateButton("Link Controller");
            linkControllerButton.button.onClick.AddListener(() =>
            {
                StartCoroutine("LinkController");
            });

            CreateTextField(new JSONStorableString("HardLink", "Hard linking sets the position directly instead of using a spring. Less controlled but reduces physics lag."), true).height = 10;
            hardLink = new JSONStorableBool("Use Hard Linking", false);
            RegisterBool(hardLink);
            var hardLinkToggle = CreateToggle(hardLink, true);

            linkJointSpring = new JSONStorableFloat("Link Joint Spring", 100000f, 0f, 100000f);
            var jointSpringSlider = AddSlider(linkJointSpring, (float val) => {
                railJointDrive.positionSpring = val;
                if(railJoint != null){
                    railJoint.xDrive = railJointDrive;
                    railJoint.yDrive = railJointDrive;
                    railJoint.zDrive = railJointDrive;
                }
            }, true);

            linkMaxForce = new JSONStorableFloat("Link Max Force", 100000f, 0f, 100000f);
            var linkMaxForceSlider = AddSlider(linkMaxForce, (float val) => {
                railJointDrive.maximumForce = val;
                if(railJoint != null){
                    railJoint.xDrive = railJointDrive;
                    railJoint.yDrive = railJointDrive;
                    railJoint.zDrive = railJointDrive;
                }
            }, true);

            maxVelocity = new JSONStorableFloat("Max Velocity", 10f, 0.01f, 10f);
            var maxVelocitySlider = AddSlider(maxVelocity, null, true);

            existingLinkType = new JSONStorableFloat("LinkType", LinkType.NONE, 0f, 100f);
            RegisterFloat(existingLinkType);
		}

        private void Start(){
            if(existingLinkType.val == LinkType.CONTROLLER) {
                StartCoroutine("LinkController");
            } else if(existingLinkType.val == LinkType.RIGIDBODY) {
                StartCoroutine("LinkBody");
            }
        }

        private bool ValidateSelection(){

            var animationAtom = SuperController.singleton.GetAtomByUid(selectedAnimationPattern);
            if(animationAtom == null){
                SuperController.LogMessage($"Could not find animation pattern {selectedAnimationPattern}");
                return false;
            }
            animationPattern = animationAtom.GetComponentInChildren<AnimationPattern>();

            if(animationPattern == null){
                SuperController.LogMessage($"Error loading animation pattern {animationPattern}");
                return false;
            }

            controller = controllers[selectedController];
            if(controller == null){
                SuperController.LogMessage($"Could not find controller {selectedController}");
                return false;
            }

            return true;
        }

        private IEnumerator LinkBody(){
            Unlink();

            if(!ValidateSelection())
                yield break;

            yield return CreateRailDriver();

            var railRB = railDriver.GetComponentInChildren<Rigidbody>();
            railDriverController = railDriver.GetComponentInChildren<FreeControllerV3>();
            
            int closestCurveIndex;
            float closestCurveT;
            linkedRB = controller.followWhenOffRB;
            Vector3 nearest = NearestPointOnBezierCurve(animationPattern, controller.transform.position, previousClosestCurvePosition, 1000f, out closestCurveIndex, out closestCurveT);

            previousClosestCurvePosition = closestCurveIndex + closestCurveT;

            railDriverController.transform.position = nearest;
            
            railDriverController.transform.rotation = Quaternion.identity;
            //railDriverController.transform.rotation = linkedRB.transform.rotation;
            
            railJoint = linkedRB.gameObject.AddComponent<ConfigurableJoint>();
            if(linkRotation.val){
                linkedRB.transform.rotation = Quaternion.identity;
                yield return new WaitForFixedUpdate();
            }
            createdObjects.Add(railJoint);
            
            railJoint.autoConfigureConnectedAnchor = false;
			railJoint.connectedBody = railRB;
			railJoint.anchor = Vector3.zero;
			railJoint.connectedAnchor = Vector3.zero;
            railJoint.axis =  linkedRB.transform.rotation.eulerAngles;

            railJoint.angularXMotion = ConfigurableJointMotion.Free;
			railJoint.angularYMotion = ConfigurableJointMotion.Free;
			railJoint.angularZMotion = ConfigurableJointMotion.Free;

            railJointDrive.positionSpring = linkJointSpring.val;
            railJointDrive.maximumForce = linkMaxForce.val;
			
			railJoint.xDrive = railJointDrive;
			railJoint.yDrive = railJointDrive;
			railJoint.zDrive = railJointDrive;

            if(linkRotation.val){
                
                railJoint.angularXMotion = ConfigurableJointMotion.Locked;
			    railJoint.angularYMotion = ConfigurableJointMotion.Locked;
			    railJoint.angularZMotion = ConfigurableJointMotion.Locked;

                linkedRB.transform.rotation = railDriverController.transform.rotation;
                rotationLinked = true;
            }

            existingLinkType.val = LinkType.RIGIDBODY;

            yield break;
        }

        private IEnumerator LinkController(){
            Unlink();

            if(!ValidateSelection())
                yield break;


            linkedRB = controller.followWhenOffRB;
            var controllerRB = controller.GetComponent<Rigidbody>();

            foreach(var joint in linkedRB.GetComponentsInChildren<ConfigurableJoint>()){
                
                if(joint.connectedBody == controllerRB){
                    controllerJoint = joint;
                }
            }

         
            if(controllerJoint == null){
                SuperController.LogMessage($"Unable to find control joint for {linkedRB}");
            }

            existingLinkType.val = LinkType.CONTROLLER;

            if(linkRotation.val){
                rotationLinked = true;
            }

            yield break;
        }

        private void Unlink(){
            
            if(railJoint != null) {
                UnityEngine.Object.Destroy(railJoint);
                railJoint = null;
            }
            existingLinkType.val = LinkType.NONE;
            rotationLinked = false;
        }

        private void FixedUpdate(){
            try{
            if(existingLinkType.val != LinkType.NONE) {

                float dTime = Time.fixedDeltaTime;
                int closestCurveIndex = (int)previousClosestCurvePosition;
                float closestCurveT = previousClosestCurvePosition - closestCurveIndex;
                
                Vector3 nearest;
                if(existingLinkType.val == LinkType.CONTROLLER) {
                    Vector3 anchorOffset = Quaternion.LookRotation(linkedRB.transform.forward)*controllerJoint.anchor;
                    nearest = NearestPointOnBezierCurve(animationPattern, linkedRB.transform.position + anchorOffset, previousClosestCurvePosition, maxVelocity.val*dTime, out closestCurveIndex, out closestCurveT);
                    controller.transform.position = nearest;


                } else if(existingLinkType.val == LinkType.RIGIDBODY) {
                    nearest = NearestPointOnBezierCurve(animationPattern, controller.transform.position, previousClosestCurvePosition, maxVelocity.val*dTime, out closestCurveIndex, out closestCurveT);
                    
                    railDriverController.transform.position = nearest;
                    
                    if(hardLink.val){
                        linkedRB.transform.position = nearest;
                    }
                }

                if(rotationLinked){
                    Quaternion targetRotation;
                    if(closestCurveIndex == animationPattern.steps.Length-1){
                        targetRotation = animationPattern.steps[closestCurveIndex].transform.rotation;
                    } else if(closestCurveIndex == 0 && animationPattern.loop){
                        Quaternion rotationA = animationPattern.steps[animationPattern.steps.Length-1].transform.rotation;
                        Quaternion rotationB = animationPattern.steps[closestCurveIndex+1].transform.rotation;
                        targetRotation = Quaternion.Lerp(rotationA, rotationB, closestCurveT);
                    } else {
                        Quaternion rotationA = animationPattern.steps[closestCurveIndex].transform.rotation;
                        Quaternion rotationB = animationPattern.steps[closestCurveIndex+1].transform.rotation;
                        targetRotation = Quaternion.Lerp(rotationA, rotationB, closestCurveT);
                    }

                    if(existingLinkType.val == LinkType.CONTROLLER) {
                        controller.transform.rotation = targetRotation;
                    } else if(existingLinkType.val == LinkType.RIGIDBODY) {
                        railDriverController.transform.rotation = targetRotation;
                    }
                }

                previousClosestCurvePosition = closestCurveIndex + closestCurveT;
            }
            } catch (Exception e){
                SuperController.LogMessage($"{e}");
            }
        }

        private IEnumerator CreateRailDriver() {
			string railDriverUid = $"{containingAtom.uid}#{storeId}#RailDriver";
			
			railDriver = SuperController.singleton.GetAtomByUid(railDriverUid);
			if(railDriver == null){
				yield return SuperController.singleton.AddAtomByType("Empty", railDriverUid);
				railDriver = SuperController.singleton.GetAtomByUid(railDriverUid);
			}	
			
			railDriver.hidden = true;
		}   
		
	
		private List<string> GetAnimationPatterns() {
            return SuperController.singleton.GetAtoms().Where(a => a.category == "Animation" && a.GetComponentInChildren<AnimationPattern>() != null).Select(a => a.uid).ToList();
		}
		
		private void OnDisable(){
         
		}
		
		private void OnDestroy(){
		    if(railDriver != null){
				SuperController.singleton.RemoveAtom(railDriver);
			}
			foreach(var obj in createdObjects){
				UnityEngine.Object.Destroy(obj);
			}
		}
		
		public UIDynamicSlider AddSlider(JSONStorableFloat storableFloat, UnityAction<float> onChange = null, bool rightSide = false)
		{
			UIDynamicSlider uiSlider = this.CreateSlider(storableFloat, rightSide);
			uiSlider.rangeAdjustEnabled = true;
			
			if(onChange != null){
				uiSlider.slider.onValueChanged.AddListener(onChange);
			}
			
			return uiSlider;
		}
		
		public Vector3 NearestPointOnBezierCurve(CubicBezierCurve curve, Vector3 pnt, float previousClosestPosition, float distanceLimit, out int closestSegment, out float closestT) {
			
			float closestDistSquared = 1000f;
			closestT = previousClosestPosition;
			Vector3 pointOnCurve = Vector3.zero;
			Vector3 closestPoint = Vector3.zero;

            float numSegments = (float)(curve.points.Length);
            float prevClosestT = numSegments/2f;

            for(float r = numSegments/2f; r >= 0.001f; r *=0.1f){
                float step = r/numSegments/5f;
                float maxCurveDistance = Math.Max(step, distanceLimit);
				for(float t = Math.Max(prevClosestT-r, 0f); t < Math.Min(prevClosestT+r, numSegments); t += step){
                    int segment = (int)t;
                    float offset = t - segment;
                    pointOnCurve = curve.GetPositionFromPoint(segment, offset);
					float distSquared = (pnt-pointOnCurve).sqrMagnitude;

                    
                    float curveDistance = Math.Abs(previousClosestPosition - t);
                    // Handle loop-around
                    curveDistance = Math.Min(curveDistance, (numSegments - t) + previousClosestPosition); 
                    curveDistance = Math.Min(curveDistance, (numSegments - previousClosestPosition) + t); 

					if(distSquared < closestDistSquared && curveDistance < distanceLimit){
						closestT = t;
						closestDistSquared = distSquared;
						closestPoint = pointOnCurve;
					}
				}
                prevClosestT = closestT;
			}

            closestSegment = (int) closestT;
			closestT = closestT - closestSegment;

            return closestPoint;
		}
	}
}